﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Threading;

using Microsoft.Lync.Model;
using Microsoft.Lync.Model.Room;

namespace LyncPersistentChat
{
    public partial class MainWindow : Window, INotifyPropertyChanged
    {
        private LyncClient _client = null;
        private Dictionary<string, Room> _followedRoomsList = new Dictionary<string, Room>();
        private ObservableCollection<string> eventMessages = new ObservableCollection<string>();
        public event PropertyChangedEventHandler PropertyChanged;

        public MainWindow()
        {
            InitializeComponent();
            this.DataContext = this;

            GetLyncClient();
        }

        #region Initialization

        /// <summary>
        /// Gets the LyncClient instance that encapsulates the Lync client platform and registers
        /// for events on both LyncClient and the RoomManager. 
        /// </summary>
        void GetLyncClient()
        {
            try
            {
                //TODO: 4.1.1 - Set up the Lync client and subscribe to changes in its state
                // Set up the Lync client and subscribe to changes in its state
                _client = LyncClient.GetClient();
                _client.StateChanged += Client_StateChanged;

                // Subscribe to RoomManagerEvents
                _client.RoomManager.FollowedRoomAdded += RoomManager_FollowedRoomAdded;
                _client.RoomManager.FollowedRoomRemoved += RoomManager_FollowedRoomRemoved;
                _client.RoomManager.RoomManagerStateChanged += RoomManager_RoomManagerStateChanged;


                lstRoomEvents.DataContext = eventMessages;

                LoadInitialFollowedRoomList();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        void Client_StateChanged(object sender, ClientStateChangedEventArgs e)
        {
            OnPropertyChanged("SelfStatus");
            OnPropertyChanged("SelfDisplayName");

            // if client is signing back in, reinit the app
            if (e.NewState == ClientState.SignedIn)
            {
                Dispatcher.BeginInvoke(
                    DispatcherPriority.Normal,
                    (Action)delegate()
                    {
                        OnPropertyChanged("SelectedRoomDescription");
                        OnPropertyChanged("SelectedRoomType");
                        OnPropertyChanged("SelectedRoom");
                    });

            }
            else if (e.NewState == ClientState.SignedOut)
            {
                ClearContent();
                OnPropertyChanged("SelectedRoomDescription");
                OnPropertyChanged("SelectedRoomType");
                OnPropertyChanged("SelectedRoom");
            }
        }

        void RoomManager_FollowedRoomAdded(object sender, FollowedRoomsChangedEventArgs e)
        {
            if (_client.State != ClientState.SignedIn) return;

            // add to our internal dictionary and the UI list
            string roomTitle = e.Room.Properties[RoomProperty.Title].ToString();

            if (!_followedRoomsList.ContainsKey(roomTitle))
            {
                _followedRoomsList.Add(roomTitle, e.Room);

                LogEvent(e.Room, string.Format("Adding room {0} to list", roomTitle));

                Dispatcher.BeginInvoke(
                    DispatcherPriority.Normal,
                    (Action)delegate()
                    {
                        lstFollowedRooms.Items.Add(roomTitle);
                    });
            }
        }

        void RoomManager_FollowedRoomRemoved(object sender, FollowedRoomsChangedEventArgs e)
        {
            if (_client.State != ClientState.SignedIn) return;

            // remove from our internal dictionary and the UI list
            string roomTitle = e.Room.Properties[RoomProperty.Title].ToString();

            _followedRoomsList.Remove(roomTitle);

            LogEvent(e.Room, string.Format("Removing room {0} from list", roomTitle));

            Dispatcher.BeginInvoke(
                DispatcherPriority.Normal,
                (Action)delegate()
                {
                    lstFollowedRooms.Items.Remove(roomTitle);
                });
        }

        void RoomManager_RoomManagerStateChanged(object sender, RoomManagerStateChangedEventArgs e)
        {
            if (_client.State != ClientState.SignedIn) return;

            LogEvent(null, string.Format("Room manager state changed to {0}", e.NewState.ToString()));
        }

        /// <summary>
        /// Loads a UI list box with the room title strings of all group chat rooms that
        /// the user has "followed"... added to the contact list.
        /// These followed rooms are locally cached by the Lync client platform.
        /// </summary>
        private void LoadInitialFollowedRoomList()
        {
            if (_client.State != ClientState.SignedIn) return;
            try
            {
                //TODO: 4.1.4 - Iterate through FollowedRooms and add them to application list
                // Iterate on the collection of Room instances that are followed by the user
                foreach (Room followedRoom in _client.RoomManager.FollowedRooms)
                {
                    // Add a room to the Dictionary<string,Microsoft.Lync.Model.Room.Room>() 
                    // Dictionary entry key is room title and value is room instance
                    _followedRoomsList.Add(followedRoom.Properties[RoomProperty.Title].ToString(), followedRoom);

                    // Add a followed room title to the UI list box of followed room titles.
                    // User selects a room title from the list box and the corresponding Room instance is    obtained from the _followedRoomsList dictonary.
                    lstFollowedRooms.Items.Add(followedRoom.Properties[RoomProperty.Title].ToString());

                }


                lstFollowedRooms.Focus();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        #endregion

        #region Shutdown

        private void ClearContent()
        {
            _followedRoomsList.Clear();

            Dispatcher.Invoke(
                DispatcherPriority.Normal,
                (Action)delegate()
                {
                    lstFollowedRooms.Items.Clear();
                    lstRoomRoster.Children.Clear();
                    lstRoomMessages.Children.Clear();
                });
        }

        #endregion

        #region UI Binding Properties

        public Room SelectedRoom { get; set; }

        public string SelectedRoomDescription
        {
            get
            {
                if (SelectedRoom != null && SelectedRoom.Properties != null)
                {
                    return (string)SelectedRoom.Properties[RoomProperty.Description];
                }
                else
                {
                    return "";
                }
            }
        }

        public RoomType SelectedRoomType
        {
            get
            {
                if (SelectedRoom != null && SelectedRoom.Properties != null)
                {
                    return (RoomType)SelectedRoom.Properties[RoomProperty.Type];
                }
                else
                {
                    return RoomType.Invalid;
                }
            }
        }

        public string SelfDisplayName
        {
            get
            {
                if (_client != null && _client.State == ClientState.SignedIn)
                {
                    return _client.ContactManager
                                .GetContactByUri(_client.Self.Contact.Uri)
                                .GetContactInformation(ContactInformationType.DisplayName).ToString();
                }
                else
                {
                    return "";
                }
            }
        }

        public string SelfStatus
        {
            get
            {
                if (_client != null)
                {
                    return _client.State.ToString();
                }
                else
                {
                    return "";
                }
            }
        }

        private void OnPropertyChanged(string name)
        {
            if (PropertyChanged != null)
            {
                PropertyChanged(this, new PropertyChangedEventArgs(name));
            }
        }

        #endregion

        #region UI Events

        void lstRoomEvents_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            if (SelectedRoom != null)
            {
                UnsubscribeFromFollowedRoomEvents(SelectedRoom);
                ClearLogEvents();
                ClearMessages();
            }

            Room stage = null;

            if ((sender as ListBox).SelectedValue != null)
            {
                var title = (sender as ListBox).SelectedValue.ToString();

                //TODO: 4.1.5 - Retrieve the ListItem selected room
                stage = _client.RoomManager.FollowedRooms.Where(p => p.Properties[RoomProperty.Title].Equals(title)).FirstOrDefault();
            }

            if (stage != null)
            {
                SelectedRoom = stage;

                LogEvent(SelectedRoom, "Subscribing to events for room: ");
                UpdateParticipantList(SelectedRoom.Participants);

                //if a selected room is found subscribe to its events
                SubscribeToFollowedRoomEvents(SelectedRoom);

                // signal that we want to be able to filter messages before they are sent
                if (!SelectedRoom.IsOutgoingMessageFilterEnabled)
                {
                    //TODO: 4.1.10 - Enable outgoing message filtering
                    SelectedRoom.EnableOutgoingMessageFilter();
                }

                // change properties for data binding            
                OnPropertyChanged("SelectedRoom");
                OnPropertyChanged("SelectedRoomDescription");
                OnPropertyChanged("SelectedRoomType");
                OnPropertyChanged("SelfDisplayName");
                OnPropertyChanged("SelfStatus");
            }

        }

        /// <summary>
        /// Sends a message to a selected chat room whenever the user presses enter.
        /// </summary>
        private void OnKeyUp(object sender, KeyEventArgs e)
        {
            // If the key pressed is enter, submit
            if (e.Key == Key.Enter)
            {
                // Make sure there is a room selected
                if (SelectedRoom != null)
                {
                    SendMessageToRoom(SelectedRoom, txtMessage.Text);
                }
                else
                {
                    txtMessageStatus.Text = "Please select a room.";
                }
            }
        }

        #endregion

        #region Lync Client Events

        void SubscribeToFollowedRoomEvents(Room followedRoom)
        {
            //TODO: 4.1.2 - Subscribe to FollowedRoom events
            followedRoom.IsSendingMessage += followedRoom_IsSendingMessage;
            followedRoom.JoinStateChanged += followedRoom_JoinStateChanged;
            followedRoom.MessagesReceived += followedRoom_MessagesReceived;
            followedRoom.ParticipantAdded += followedRoom_ParticipantAdded;
            followedRoom.ParticipantRemoved += followedRoom_ParticipantRemoved;
            followedRoom.PropertyChanged += followedRoom_PropertyChanged;
            followedRoom.UnreadMessageCountChanged += followedRoom_UnreadMessageCountChanged;


            LogEvent(SelectedRoom, "Subscribed to IsSendingMessage event");
            LogEvent(SelectedRoom, "Subscribed to JoinStateChanged event");
            LogEvent(SelectedRoom, "Subscribed to MessageReceived event");
            LogEvent(SelectedRoom, "Subscribed to ParticipantAdded event");
            LogEvent(SelectedRoom, "Subscribed to ParticipantRemoved event");
            LogEvent(SelectedRoom, "Subscribed to PropertyChanged event");
            LogEvent(SelectedRoom, "Subscribed to UnreadMessageCountChanged event");

        }

        void UnsubscribeFromFollowedRoomEvents(Room followedRoom)
        {
            //TODO: 4.1.3 - Unsubscribe to FollowedRoom events
            followedRoom.IsSendingMessage -= followedRoom_IsSendingMessage;
            followedRoom.JoinStateChanged -= followedRoom_JoinStateChanged;
            followedRoom.MessagesReceived -= followedRoom_MessagesReceived;
            followedRoom.ParticipantAdded -= followedRoom_ParticipantAdded;
            followedRoom.ParticipantRemoved -= followedRoom_ParticipantRemoved;
            followedRoom.PropertyChanged -= followedRoom_PropertyChanged;
            followedRoom.UnreadMessageCountChanged -= followedRoom_UnreadMessageCountChanged;

        }

        void followedRoom_UnreadMessageCountChanged(object sender, UnreadMessageCountChangedEventArgs e)
        {
            OnPropertyChanged("SelectedRoom");
        }

        void followedRoom_PropertyChanged(object sender, RoomPropertyChangedEventArgs e)
        {
            LogEvent(sender as Room, "Property: " + e.Property.ToString() + " changed to value: " + e.Value.ToString());
            OnPropertyChanged("SelectedRoom");
        }

        void followedRoom_ParticipantRemoved(object sender, RoomParticipantsEventArgs e)
        {
            LogEvent(sender as Room, "Participant removed");

            this.Dispatcher.BeginInvoke(
                DispatcherPriority.Normal,
                (Action)delegate()
                {
                    UpdateParticipantList(SelectedRoom.Participants);
                });
        }

        void followedRoom_ParticipantAdded(object sender, RoomParticipantsEventArgs e)
        {
            LogEvent(sender as Room, "Participant added.");

            this.Dispatcher.BeginInvoke(
                DispatcherPriority.Normal,
                (Action)delegate()
                {
                    UpdateParticipantList(SelectedRoom.Participants);
                });
        }

        void followedRoom_MessagesReceived(object sender, RoomMessagesEventArgs e)
        {
            LogEvent(sender as Room, "Message Received ");

            this.Dispatcher.BeginInvoke(
                DispatcherPriority.Normal,
                (Action)delegate()
                {
                    var orderedMessages = e.Messages.OrderBy(m => m.SentTime);

                    foreach (var message in orderedMessages)
                    {
                        TextBlock stage = new TextBlock();
                        stage.Style = this.FindResource("EventMsgStyle") as Style;
                        stage.Inlines.Add(new Italic(new Run(message.SentTime.ToString() + "  ")));
                        stage.Inlines.Add(new Bold(new Run(
                            _client.ContactManager.GetContactByUri(message.SenderUri)
                            .GetContactInformation(ContactInformationType.DisplayName).ToString() + ": ")));
                        stage.Inlines.Add((string)message.MessageDictionary[RoomMessageFormat.PlainText]);

                        lstRoomMessages.Children.Insert(0, stage);
                    }
                });
        }

        void followedRoom_JoinStateChanged(object sender, RoomJoinStateChangedEventArgs e)
        {
            LogEvent(sender as Room, "Joined State Changed: " + e.JoinState);
        }

        void followedRoom_IsSendingMessage(object sender, RoomMessageEventArgs e)
        {
            LogEvent(sender as Room, "Sending Message ");

            //TODO: 4.1.11 - Instantiate RoomMessage from RoomMessageEventArgs
            RoomMessage filteredMessage = e.Message;

            string filterText = "";
            this.Dispatcher.Invoke(
                DispatcherPriority.Normal,
                (Action)delegate()
                {
                    filterText = txtFilter.Text;
                });

            //TODO: 4.1.12 - Filter the RoomMessage instance
            if (!string.IsNullOrEmpty(filterText)
                && ((string)filteredMessage.MessageDictionary[RoomMessageFormat.PlainText]).IndexOf(filterText) >= 0)
            {
                var rtBox = new System.Windows.Forms.RichTextBox();
                rtBox.Text = (string)filteredMessage.MessageDictionary[RoomMessageFormat.PlainText];
                rtBox.ForeColor = System.Drawing.Color.Red;

                filteredMessage.MessageDictionary[RoomMessageFormat.Rtf] = rtBox.Rtf;
                SelectedRoom.SendFilteredMessage(filteredMessage, RoomMessageFilteringAction.Replaced);
            }
            else
            {
                // send unaltered message
                SelectedRoom.SendFilteredMessage(filteredMessage, RoomMessageFilteringAction.Passed);
            }

        }

        #endregion

        #region Lync Client Messaging

        /// <summary>
        /// Sends a text message to a group chat room using the Room API.
        /// </summary>
        private void SendMessageToRoom(Room room, string message)
        {
            if (!string.IsNullOrEmpty(message))
            {
                bool isChecked = false;
                //TODO: 4.1.7 - Create instance of RoomMessageType enum
                RoomMessageType messageType = RoomMessageType.Regular;

                // get the value of the alert checkbox
                Dispatcher.Invoke(
                    DispatcherPriority.Normal,
                    (Action)delegate()
                    {
                        if (chkboxAlert.IsChecked.HasValue)
                        {
                            isChecked = chkboxAlert.IsChecked.Value;
                        }
                    });

                // change alert type
                if (isChecked)
                {
                    //TODO: 4.1.8 - Set the MessageType instance to Alert
                    messageType = RoomMessageType.Alert;
                }

                try
                {
                    //TODO: 4.1.9 - Send a message to the active room
                    // Start sending message using an asynchronous call and a lambda callback
                    room.BeginSendMessage(message, messageType, (ar) =>
                    {
                        string statusMessage = "";
                        try
                        {
                            // Pass the result of the call to the End method to complete the send.
                            room.EndSendMessage(ar);

                            // Change text to show success.
                            statusMessage = "Message sent.";
                        }
                        catch
                        {
                            // Change text to show failure.
                            statusMessage = "Message failed to send.";
                        }
                        finally
                        {
                            // Clear text box text on UI thread.
                            Dispatcher.Invoke(
                                  DispatcherPriority.Normal,
                                  (Action)delegate()
                                  {
                                      txtMessage.Text = "";
                                      txtMessageStatus.Text = statusMessage;
                                  });
                        }
                    }, null);

                }
                catch (NotSignedInException)
                {
                    Dispatcher.Invoke(
                        DispatcherPriority.Normal,
                        (Action)delegate()
                        {
                            txtMessageStatus.Text = "Not signed in.";
                        });
                }
                catch
                {
                    throw;
                }
            }
        }

        #endregion

        #region Participants

        /// <summary>
        /// Fills a list box with the display names of active room participants.
        /// </summary>
        /// <param name="users"></param>
        private void UpdateParticipantList(IList<RoomUser> users)
        {
            string displayName = string.Empty;
            lstRoomRoster.Children.Clear();

            // Iterate on list of room users
            foreach (RoomUser roomUser in users)
            {
                // Get the display name of a room user
                if (roomUser.Uri != null)
                {
                    //TODO: 4.1.6 - Retrieve participants FullName
                    displayName = _client.ContactManager
                        .GetContactByUri(roomUser.Uri)
                        .GetContactInformation(ContactInformationType.DisplayName).ToString();

                    var txtDisplayName = new TextBlock()
                    {
                        Text = displayName,
                        Style = this.FindResource("EventMsgStyle") as Style
                    };
                    // Add the display name to the list box if it is not already in the list box
                    if (!lstRoomRoster.Children.Contains(txtDisplayName))
                    {
                        lstRoomRoster.Children.Add(txtDisplayName);
                    }
                }
            }
        }

        #endregion

        #region Logging

        void LogEvent(Room followedRoom, string message)
        {
            this.Dispatcher.BeginInvoke(
                DispatcherPriority.Normal,
                (Action)delegate()
                {
                    string space = "  ";

                    if (lstRoomEvents.Children.Count == 0 && followedRoom != null)
                    {
                        TextBlock stage = new TextBlock();
                        stage.Style = this.FindResource("EventMsgStyle") as Style;
                        stage.Inlines.Add(new Italic(new Run(DateTime.Now.ToString() + space)));
                        stage.Inlines.Add(message);
                        stage.Inlines.Add(new Bold(new Run(space + followedRoom.Properties[RoomProperty.Title].ToString())));

                        lstRoomEvents.Children.Insert(0, stage);

                    }
                    else
                    {
                        TextBlock stage = new TextBlock();
                        stage.Style = this.FindResource("EventMsgStyle") as Style;
                        stage.Inlines.Add(new Italic(new Run(DateTime.Now.ToString() + space)));
                        stage.Inlines.Add(message);

                        lstRoomEvents.Children.Insert(0, stage);
                    }
                }
            );
        }

        void ClearLogEvents()
        {
            this.Dispatcher.BeginInvoke(
                DispatcherPriority.Normal,
                (Action)delegate()
                {
                    lstRoomEvents.Children.Clear();
                }
            );
        }

        void ClearMessages()
        {
            this.Dispatcher.BeginInvoke(
                DispatcherPriority.Normal,
                (Action)delegate()
                {
                    lstRoomMessages.Children.Clear();
                }
            );
        }

        #endregion
    }

}