﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Configuration;
using System.Linq;
using System.Security.Cryptography.X509Certificates;
using System.Threading;
using System.Xml;

using Microsoft.Rtc.Collaboration;
using Microsoft.Rtc.Collaboration.PersistentChat;
using Microsoft.Rtc.Collaboration.PersistentChat.Management;
using Microsoft.Rtc.Signaling;

namespace LyncPersistentChatServerSdkBot
{
    class ServerSDKBot
    {
        CollaborationPlatform _platform;
        LocalEndpoint _userEndpoint;
        PersistentChatEndpoint _persistentChatEndpoint;
        ChatRoomSession _chatRoomSession;

        Timer _messageTimer; 
        object _syncObject = new object();
        int messageNum = 0;
        const int maxMessages = 4;

        private static string _applicationId;
        private static string _applicationName;
        private static string _primaryLabUserId;
        private static string _persistentChatEndpointUri;
        private static string _persistentChatAddInUri;
        private static string _roomNamePrefix;

        public void Start()
        {
            if (String.IsNullOrEmpty(ConfigurationManager.AppSettings["ApplicationId"]) ||
                (String.IsNullOrEmpty(ConfigurationManager.AppSettings["ApplicationName"])) ||
                (String.IsNullOrEmpty(ConfigurationManager.AppSettings["PrimaryLabUserId"])) ||
                (String.IsNullOrEmpty(ConfigurationManager.AppSettings["PersistentChatEndpointUri"])) ||
                (String.IsNullOrEmpty(ConfigurationManager.AppSettings["PersistentChatAddInUri"])) ||
                (String.IsNullOrEmpty(ConfigurationManager.AppSettings["RoomNamePrefix"])))
            {
                throw new ArgumentNullException("Please provide values for ApplicationId, ApplicationName, PrimaryLabUserId, PersistentChatEndpointUri, PersistentChatAddInUri, and RoomNamePrefix in app.config.");
            }

            _applicationId = ConfigurationManager.AppSettings["ApplicationId"];
            _applicationName = ConfigurationManager.AppSettings["ApplicationName"];
            _primaryLabUserId = ConfigurationManager.AppSettings["PrimaryLabUserId"];
            _persistentChatEndpointUri = ConfigurationManager.AppSettings["PersistentChatEndpointUri"];
            _persistentChatAddInUri = ConfigurationManager.AppSettings["PersistentChatAddInUri"];
            _roomNamePrefix = ConfigurationManager.AppSettings["RoomNamePrefix"];

            var platformSettings = new ProvisionedApplicationPlatformSettings(_applicationName, _applicationId);
            _platform = new CollaborationPlatform(platformSettings);

            try
            {
                _platform.BeginStartup(startupAsyncResult =>
                    {
                        try
                        {
                            _platform.EndStartup(startupAsyncResult);
                            Console.WriteLine("Started platform.");
                            EstablishEndpoint();
                        }
                        catch (RealTimeException ex)
                        {
                            Console.WriteLine(ex);
                        }
                    },
                    null);
            }
            catch (InvalidOperationException ex)
            {
                Console.WriteLine(ex);
            }
        }

        private void EstablishEndpoint()
        {
            var endpointSettings = new UserEndpointSettings(_primaryLabUserId);
            _userEndpoint = new UserEndpoint(_platform, endpointSettings);

            try
            {
                _userEndpoint.BeginEstablish(establishAsyncResult =>
                {
                    try
                    {
                        _userEndpoint.EndEstablish(establishAsyncResult);
                        Console.WriteLine("User endpoint established.");
                        EstablishPersistentChatEndpoint();
                    }
                    catch (RealTimeException ex)
                    {
                        Console.WriteLine(ex);
                    }
                },
                null);
            }
            catch (InvalidOperationException ex)
            {
                Console.WriteLine(ex);
            }
        }

        private void EstablishPersistentChatEndpoint()
        {
            //TODO: 4.3.1 - Create a new Persistent Chat endpoint
            

            try
            {
                _persistentChatEndpoint.BeginEstablish(
                    establishAsyncResult =>
                    {
                        try
                        {
                            _persistentChatEndpoint.EndEstablish(establishAsyncResult);
                            Console.WriteLine("Persistent Chat endpoint established.");
                            CreateChatRoom(_roomNamePrefix + "_" + Guid.NewGuid().ToString());
                        }
                        catch (RealTimeException ex)
                        {
                            Console.WriteLine(ex);
                        }

                    }, null);
            }
            catch (InvalidOperationException ex)
            {
                Console.WriteLine(ex);
            }
        }

        private void CreateChatRoom(string roomName)
        {
            _persistentChatEndpoint
                .PersistentChatServices
                .ChatRoomManagementServices
                .BeginFindCategoriesWithCreateRights(ar =>
                {
                    try
                    {
                        ReadOnlyCollection<ChatRoomCategorySummary> categories = null;

                        //TODO: 4.3.2 - Find a category to create a room in


                        if (categories.Count > 0)
                        {
                            _persistentChatEndpoint
                                .PersistentChatServices
                                .ChatRoomManagementServices
                                .BeginCreateChatRoom(
                                    roomName,
                                    categories[0].Uri,
                                    arCreateRoom =>
                                    {
                                        try
                                        {
                                            Uri newRoomUri = null;
                                            
                                            //TODO: 4.3.3 - Create a room
                                            

                                            // Add user as member
                                            AddUserAsMemberToRoom(roomName, newRoomUri);
                                        }
                                        catch (Exception ex)
                                        {
                                            Console.WriteLine(ex);
                                        }
                                    }, null);
                        }
                        else
                        {
                            throw new Exception("Could not find any categories to create a room in.");
                        }
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine(ex);
                    }
                }, null);
        }

        private void AddUserAsMemberToRoom(string chatRoomName, Uri roomUri)
        {
            _persistentChatEndpoint
                .PersistentChatServices
                .UserAdministrationServices
                .BeginGetUser(
                    new Uri(_primaryLabUserId), 
                    ar =>
                    {
                        try
                        {
                            PersistentChatUser me = _persistentChatEndpoint
                                                .PersistentChatServices
                                                .UserAdministrationServices
                                                .EndGetUser(ar);

                            var memberUsers = new List<PersistentChatPrincipalSummary> { me };

                            _persistentChatEndpoint
                                .PersistentChatServices
                                .ChatRoomManagementServices
                                .BeginAddUsersOrGroupsToRole(
                                    ChatRoomRole.Member, 
                                    roomUri, 
                                    memberUsers, 
                                    asr =>
                                    {
                                        try
                                        {
                                            //TODO: 4.3.4 - Add the user as a member of the room
                                            

                                            // Join room after becoming a member
                                            JoinChatRoom(chatRoomName);
                                        }
                                        catch (Exception ex)
                                        {
                                            Console.WriteLine(ex);
                                        }

                                    }, null);
                        }
                        catch (Exception ex)
                        {
                            Console.WriteLine(ex);
                        }
                    }, null);
                }

        private void JoinChatRoom(string chatRoomName)
        {
            //TODO: 4.3.5 - Join the room
            
        }

        private void AssociateAddInToRoom()
        {
            _persistentChatEndpoint
                .PersistentChatServices
                .ChatRoomManagementServices
                .BeginGetChatRoom(
                    _chatRoomSession.Name, 
                    ar =>
                    {
                        try
                        {
                            var room = _persistentChatEndpoint
                                        .PersistentChatServices
                                        .ChatRoomManagementServices
                                        .EndGetChatRoom(ar);

                            if (room != null)
                            {
                                // Set the room's add-in
                                SetRoomAddIn(room);
                            }
                        }
                        catch (Exception ex)
                        {
                            Console.WriteLine(ex);
                        }
                    }, null);
        }

        private void SetRoomAddIn(ChatRoom room)
        {
            _persistentChatEndpoint
                .PersistentChatServices
                .ChatRoomManagementServices
                .BeginGetAllAddIns(
                    ar =>
                    {
                        try
                        {
                            ReadOnlyCollection<PersistentChatAddIn> addins = null;
                            PersistentChatAddIn myAddin = null;

                            //TODO: 4.3.6 - Find the add-in
                            

                            if (myAddin != null)
                            {
                                var info = new ChatRoomInformation(room);
                                info.AddIn = myAddin;
                                _persistentChatEndpoint
                                    .PersistentChatServices
                                    .ChatRoomManagementServices
                                    .BeginUpdateChatRoom(
                                        info, 
                                        arUpdateChatRoom =>
                                        {
                                            try
                                            {
                                                //TODO: 4.3.7 - Associate the add-in with the room
                                                

                                                Console.WriteLine("Associated add-in at {0} with the room.", 
                                                    info.AddIn.UriString);
                                            }
                                            catch (Exception ex)
                                            {
                                                Console.WriteLine(ex);
                                            }
                                        }, null);
                            }
                        }
                        catch (Exception ex)
                        {
                            Console.WriteLine(ex);
                        }
                    }, null);
        }

        void OnChatRoomSessionStateChanged(object sender, ChatRoomSessionStateChangedEventArgs e)
        {
            if (e.State == ChatRoomSessionState.Established)
            {
                StartMessageTimer();
            }
            else if (e.State == ChatRoomSessionState.Terminated)
            {
                StopMessageTimer();
            }
        }

        private void StartMessageTimer()
        {
            lock (_syncObject)
            {
                _messageTimer = new Timer(OnTimerIntervalElapsed, null, 5000, 5000);
            }
        }

        private void StopMessageTimer()
        {
            lock (_syncObject)
            {
                if (_messageTimer != null)
                {
                    _messageTimer.Dispose();
                    _messageTimer = null;
                }
            }
        }

        private void OnTimerIntervalElapsed(object state)
        {
            try
            {
                if (messageNum < maxMessages)
                {
                    XmlDocument xDoc = new XmlDocument();

                    xDoc.Load("NewsArticles.xml");

                    XmlNodeList articles = xDoc.GetElementsByTagName("Article");

                    string title = articles[messageNum].ChildNodes[0].InnerText.Trim(new char[] { ' ', '\n', '\r' });
                    string link = articles[messageNum].ChildNodes[1].InnerText.Trim(new char[] { ' ', '\n', '\r' });

                    Console.WriteLine();
                    Console.WriteLine(String.Format("Sending message: {0} - {1}", title, link));

                    //TODO: 4.3.8 - Send a message to the room

                    
                    messageNum++;
                }
                else
                {
                    _messageTimer.Dispose();
                    _messageTimer = null;
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex);
            }
        }

        public void Stop()
        {
            try
            {
                //TODO: 4.3.9 - Leave the room
               
            }
            catch (InvalidOperationException ex)
            {
                Console.WriteLine(ex);
            }
        }

        private void TerminateEndpoint()
        {
            try
            {
                _userEndpoint.BeginTerminate(terminateAsyncResult =>
                {
                    try
                    {
                        _userEndpoint.EndTerminate(terminateAsyncResult);

                        Console.WriteLine("Endpoint terminated.");

                        ShutDownPlatform();
                    }
                    catch (RealTimeException ex)
                    {
                        Console.WriteLine(ex);
                    }
                },
                null);
            }
            catch (InvalidOperationException ex)
            {
                Console.WriteLine(ex);
            }
        }

        private void ShutDownPlatform()
        {
            try
            {
                _platform.BeginShutdown(shutdownAsyncResult =>
                {
                    try
                    {
                        _platform.EndShutdown(shutdownAsyncResult);

                        Console.WriteLine("Shut down platform.");
                    }
                    catch (RealTimeException ex)
                    {
                        Console.WriteLine(ex);
                    }
                },
                    null);
            }
            catch (InvalidOperationException ex)
            {
                Console.WriteLine(ex);
            }
        }
    }
}
